import pytest

from tenuki.rank import Rank, RankLevel


def test_rank_parse_for_kyu():
    assert Rank.parse("1k") == Rank(value=1, level=RankLevel.KYU)


def test_rank_parse_for_dan():
    assert Rank.parse("1d") == Rank(value=1, level=RankLevel.DAN)


def test_rank_parse_for_pro():
    assert Rank.parse("1p") == Rank(value=1, level=RankLevel.PRO)


def test_rank_parse_when_wrong_input():
    with pytest.raises(ValueError) as e:
        Rank.parse("xyz")

    assert str(e.value) == "Invalid rank: xyz"


def test_rank_parse_when_wrong_value():
    with pytest.raises(ValueError) as e:
        Rank.parse("10d")

    assert str(e.value) == "Invalid rank: 10d"


def test_rank_parse_when_wrong_level():
    with pytest.raises(ValueError) as e:
        Rank.parse("1x")

    assert str(e.value) == "Invalid rank: 1x"


@pytest.mark.parametrize(
    "first_rank, second_rank, expected",
    [
        ("1k", "2k", False),
        ("2k", "1k", True),
        ("1d", "1k", False),
        ("1k", "1d", True),
        ("2d", "1d", False),
        ("1d", "2d", True),
        ("1p", "9d", False),
        ("9d", "1p", True),
        ("2p", "1p", False),
        ("1p", "2p", True),
    ],
)
def test_rank_lower(first_rank: str, second_rank: str, expected: bool):
    assert (Rank.parse(first_rank) < Rank.parse(second_rank)) is expected


def test_rank_str():
    assert str(Rank.parse("1k")) == "1k"
    assert str(Rank.parse("1d")) == "1d"
    assert str(Rank.parse("1p")) == "1p"


@pytest.mark.parametrize(
    "first_rank, second_rank, expected",
    [
        ("9k", "5k", -4),
        ("5k", "9k", 4),
        ("2k", "2d", -3),
        ("2d", "2k", 3),
        ("5d", "9d", -4),
        ("9d", "5d", 4),
        ("8d", "2p", -3),
        ("2p", "8d", 3),
        ("5p", "9p", -4),
        ("9p", "5p", 4),
        ("1k", "1p", -10),
        ("1p", "1k", 10),
    ],
)
def test_rank_diff(first_rank: str, second_rank: str, expected: int):
    assert Rank.parse(first_rank).diff(Rank.parse(second_rank)) == expected


@pytest.mark.parametrize(
    "first_rank, value, expected",
    [
        ("2k", 1, "1k"),
        ("2k", 2, "1d"),
        ("1k", 1, "1d"),
        ("1d", 1, "2d"),
        ("9d", 1, "1p"),
        ("1p", 1, "2p"),
        ("1k", 10, "1p"),
    ],
)
def test_rank_add_int(first_rank: str, value: int, expected: str):
    assert Rank.parse(first_rank) + value == Rank.parse(expected)


@pytest.mark.parametrize(
    "first_rank, value, expected",
    [
        ("2k", 1, "3k"),
        ("1d", 1, "1k"),
        ("2d", 1, "1d"),
        ("1p", 1, "9d"),
        ("2p", 1, "1p"),
        ("1p", 10, "1k"),
    ],
)
def test_rank_subtract_int(first_rank: str, value: int, expected: str):
    assert Rank.parse(first_rank) - value == Rank.parse(expected)



from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from functools import total_ordering


class RankLevel(Enum):
    KYU = "k"
    DAN = "d"
    PRO = "p"

    def __lt__(self, other: RankLevel) -> bool:
        order = [RankLevel.KYU, RankLevel.DAN, RankLevel.PRO]
        return order.index(self) < order.index(other)


@total_ordering
@dataclass(frozen=True)
class Rank:
    value: int
    level: RankLevel

    def __str__(self):
        return f"{self.value}{self.level.value}"

    def __post_init__(self):
        if self.value < 1 or (self.level in [RankLevel.DAN, RankLevel.PRO] and self.value > 9):
            raise ValueError(f"Invalid rank: {self.value}{self.level.value}")

    def __lt__(self, other: Rank) -> bool:
        if self.level == other.level:
            if self.level == RankLevel.KYU:
                return self.value > other.value
            else:
                return self.value < other.value
        return self.level < other.level

    def __sub__(self, value: int) -> Rank:
        return Rank._from_int(self._to_int() - value)

    def __add__(self, value: int) -> Rank:
        return Rank._from_int(self._to_int() + value)

    def diff(self, other: Rank) -> int:
        """Returns the difference between two ranks - aka 'stones' difference"""
        return self._to_int() - other._to_int()

    def _to_int(self) -> int:
        match self.level:
            case RankLevel.KYU:
                return -self.value
            case RankLevel.DAN:
                return self.value - 1
            case RankLevel.PRO:
                return self.value + 8

    @staticmethod
    def _from_int(value: int) -> Rank:
        if value < 0:
            return Rank(value=-value, level=RankLevel.KYU)
        elif value < 9:
            return Rank(value=value + 1, level=RankLevel.DAN)
        else:
            return Rank(value=value - 8, level=RankLevel.PRO)

    @staticmethod
    def parse(raw: str) -> Rank:
        """Parses a rank from a string."""
        m = re.match(r"(\d+)\s?([a-z]+)", raw.lower())
        if not m:
            raise ValueError(f"Invalid rank: {raw}")
        value, level = m.groups()
        value = int(value)
        match level:
            case "d" | "dan":
                level = RankLevel.DAN
            case "k" | "kyu":
                level = RankLevel.KYU
            case "p" | "pro":
                level = RankLevel.PRO
            case _:
                raise ValueError(f"Invalid rank: {raw}")
        if (level in [RankLevel.DAN, RankLevel.PRO] and value > 9) or value < 1:
            raise ValueError(f"Invalid rank: {raw}")
        return Rank(value=value, level=level)



import pytest

from tenuki.egf import gor_to_rank, rank_to_gor
from tenuki.rank import Rank


@pytest.mark.parametrize(
    "gor, is_pro, expected",
    [
        (2151, False, Rank.parse("2d")),
        (2150, False, Rank.parse("2d")),
        (2149, False, Rank.parse("1d")),
        (2100, False, Rank.parse("1d")),
        (2050, False, Rank.parse("1d")),
        (2049, False, Rank.parse("1k")),
        (2000, False, Rank.parse("1k")),
        (1950, False, Rank.parse("1k")),
        (1949, False, Rank.parse("2k")),
        (-900, False, Rank.parse("30k")),
        (2684, True, Rank.parse("7d")),
        (2685, True, Rank.parse("1p")),
        (2700, True, Rank.parse("1p")),
        (2714, True, Rank.parse("1p")),
        (2715, True, Rank.parse("2p")),
        (2730, True, Rank.parse("2p")),
        (3000, True, Rank.parse("9p")),
    ],
)
def test_gor_to_rank(gor: int, is_pro: bool, expected: Rank) -> None:
    assert gor_to_rank(gor=gor, is_pro=is_pro) == expected


@pytest.mark.parametrize(
    "rank, expected",
    [
        (Rank.parse("1p"), 2700),
        (Rank.parse("1d"), 2100),
        (Rank.parse("1k"), 2000),
        (Rank.parse("30k"), -900),
    ],
)
def test_rank_to_gor(rank: Rank, expected: int) -> None:
    assert rank_to_gor(rank=rank) == expected



from tenuki.rank import Rank, RankLevel


def gor_to_rank(gor: int, is_pro: bool = False) -> Rank:
    if is_pro and gor >= 2685:
        result = ((gor - 15) // 30) - 88
        return Rank(value=min(result, 9), level=RankLevel.PRO)
    if gor < 2050:
        result = -((gor - 50) // 100) + 20
        return Rank(value=result, level=RankLevel.KYU)
    else:
        result = ((gor - 50) // 100) - 19
        return Rank(value=min(result, 9), level=RankLevel.DAN)


def rank_to_gor(rank: Rank) -> int:
    match rank.level:
        case RankLevel.KYU:
            return -((rank.value - 21) * 100)
        case RankLevel.DAN:
            return (rank.value + 20) * 100
        case RankLevel.PRO:
            return (rank.value + 89) * 30



from setuptools import setup

setup(
    name="tenuki",
    version="0.2.0",
    description="A set of development tools for go/baduk/wéiqí projects.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Konrad Hałas",
    author_email="halas.konrad@gmail.com",
    url="https://github.com/konradhalas/tenuki",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.11",
    packages=["tenuki"],
    package_data={"tenuki": ["py.typed"]},
    extras_require={
        "dev": ["pytest>=7.4", "black", "mypy"]
    },
)
